;+
; NAME:
;  repwrite
; PURPOSE:
;  Update file by replacing or adding line of information
; DESCRIPTION:
; CATEGORY:
;  Utility
; CALLING SEQUENCE:
;  repwrite,file,tag,line
; INPUTS:
;
;  file - string of file name to update (or create)
;
;  tag  - key to scan at the start of the line, all lines found
;          matching this key will be replaced by the new line provided.
;          Any other line matching this tag will be removed from the file.
;
;  line - Replacement line for file.  If tag is not found, this will be
;          added to the file.  The output is left sorted by the tags.
;
; OPTIONAL INPUT PARAMETERS:
; KEYWORD INPUT PARAMETERS:
;
;  HEADER  - String array to put at start of file if file is created.
;              Default is no header.  If file already exists, then this is
;              ignored.
;
;  HEADLEN - Number of lines at the start of the file to pass over.
;              File must have at least this many lines to be valid.  If no
;              header provided, the default is 0.  If you give a header, then
;              the length of the string array is used as the default header
;              length.
;           
; OUTPUTS:
;  All output is to the file.
; KEYWORD OUTPUT PARAMETERS:
; COMMON BLOCKS:
; SIDE EFFECTS:
; RESTRICTIONS:
; PROCEDURE:
; MODIFICATION HISTORY:
;  96/11/19, Written by Marc W. Buie, Lowell Observatory
;  97/2/6, MWB, different algorithm and file is now sorted.
;  97/10/09, MWB, added HEADER and HEADLEN keywords
;  98/03/22, MWB, added NOSORT keyword
;-
pro repwrite,file,tag,line,HEADER=header,HEADLEN=headlen,NOSORT=nosort

   if badpar(file,7,0,caller='REPWRITE (file) ') then return
   if badpar(tag, 7,0,caller='REPWRITE (tag) ') then return
   if badpar(line,7,0,caller='REPWRITE (line) ') then return
   if badpar(headlen,[0,2,3],0,caller='REPWRITE (HEADLEN) ',default=-1) then return
   if badpar(header,[0,7],[0,1],caller='REPWRITE (HEADER) ',default='[[NoNe]]') then return

   IF headlen eq -1 and header[0] eq '[[NoNe]]' THEN headlen = 0
   IF headlen eq -1 and header[0] ne '[[NoNe]]' THEN headlen = n_elements(header)

   if exists(file) then begin

      ; First find the number of lines in the file
      openr,lun,file,/get_lun
      in_line=''
      nlines=0
      while(not eof(lun)) do begin
         readf,lun,in_line,format='(a1)'
         nlines=nlines+1
      endwhile
      point_lun,lun,0

      IF headlen gt nlines THEN BEGIN
         print,'repwrite: file shorter than header length, aborting'
         return
      ENDIF

      IF headlen gt 0 THEN BEGIN
         hdr=strarr(headlen)
         FOR i=0,headlen-1 DO BEGIN
            readf,lun,in_line,format='(a)'
            hdr[i] = in_line
         ENDFOR
         nlines = nlines - headlen
      ENDIF

      ; now load the data from the file
      taglen=strlen(tag)
      tags=strarr(nlines)
      info=strarr(nlines)
      for i=0,nlines-1 do begin
         readf,lun,in_line,format='(a)'
         tags[i]=strmid(in_line,0,taglen)
         info[i]=in_line
      endfor
      free_lun,lun

      ; Identify those lines that don't match the desired tag
      z = where(tag ne tags,count)

      ; save non-matching lines, add new one, and sort
      IF count ne n_elements(tags) and count ne 0 THEN BEGIN
         info = info[z]
         tags = tags[z]
      ENDIF
      IF count eq 0 THEN BEGIN
         info = line
         idx  = 0
      ENDIF ELSE BEGIN
         info = [info,line]
         tags = [tags,tag]
         IF keyword_set(nosort) THEN BEGIN
            idx=lindgen(n_elements(tags))
         ENDIF ELSE BEGIN
            idx=sort(tags)
         ENDELSE
      ENDELSE

      ; write the new ordered set back to the file.
      openw,lun,file,/get_lun
      IF headlen ne 0 THEN printf,lun,hdr,format='(a)'
      FOR i=0,n_elements(info)-1 DO BEGIN
         printf,lun,info[idx[i]]
      ENDFOR
      free_lun,lun

   endif else begin
      openw,lun,file,/get_lun
      IF header[0] ne '[[NoNe]]' THEN printf,lun,header,format='(a)'
      printf,lun,line
      free_lun,lun
   endelse

end
