;+
; NAME:
;  robomean
; PURPOSE: (one line)
;  Robust statistical moments of the data.
; DESCRIPTION:
;  This routine computes the average, average deviation, standard
;  deviation, variance, skew and kurtosis of the input data.  The various
;  output quantities are always returned as floating point scalars.
;  The statistics are computed with no regard for the dimensionality of
;  the input data.
;
;  The statistics are robust in that the data is searched for outliers
;  after the moments are computed.  If outliers are found they are
;  removed and the statistics are computed again.  This continues until
;  either no outliers are found or if the removal of outliers has an
;  insignificant effect on the statistics.
; CATEGORY:
;  Statistics
; CALLING SEQUENCE:
;  robomean,data,thresh,eps,avg,avgdev,stddev,var,skew,kurt,nfinal
; INPUTS:
;  data   - Input data to be analyzed.
;  thresh - Deviation from the mean to signify an outlier.
;  eps    - Smallest significant change in mean in units of std dev.
; OPTIONAL INPUT PARAMETERS:
;  None.
; INPUT KEYWORD PARAMETERS:
;  bad    - Array of flags, 0=good, 1=bad.  Any points removed from the sample
;              will have this flag set to 1 upon return.
; OUTPUT KEYWORD PARAMETERS:
;  STDMEAN : Optional return of standard deviation of the mean.
; OUTPUTS:
;  avg    - Sample mean.
;  avgdev - Average deviation of the data from the mean.
;  stddev - Standard deviation of the data from the mean.
;  var    - Variance of the data from the mean.
;  skew   - Skewness, third statistical moment.
;  kurt   - Kurtosis, fourth statistical moment.
;  nfinal - Number of points used in the final result.
;  new    - Vector of 'clean' pixels (optional).
; COMMON BLOCKS:
;  None.
; SIDE EFFECTS:
;  None.
; RESTRICTIONS:
;  None.
; PROCEDURE:
;  Standard techniques, see section 13.1 in Numerical Recipies.  The
;  thresh and eps values are not tremendously important.  Thresh=5.0
;  and eps=0.5 appear to work pretty well most of the time.
; MODIFICATION HISTORY:
;  Written by Marc W. Buie, Lowell Observatory, 1992 Jan 20.
;  Fix - nfinal was not being returned to calling program.
;        Doug Loucks, Lowell Observatory, 1992 Oct 22.
;  Fix - Sense of test to determine additional refinement was
;        not correct for some cases.  Changed logic to parallel
;        the C version, since the WHILE statement is now available
;        in IDL.
;  Mod - Added by Marc Buie: Argument 'new,' allowing the 'clean' pixels
;        to be returned to the caller.
;  Mod - 2/11/94, DWL, Added keyword STDMEAN to permit return of this
;        value to the caller.
;  97/03/23 - MWB, added BAD keyword and operation.  This program's operation
;        should not be any different than what it was if you don't supply BAD.
;-
PRO robomean, data, thresh, eps, avg, avgdev, stddev, var, skew, kurt, $
              nfinal,new, STDMEAN=r_stdmean, BAD=bad

savebad = n_elements(bad) eq n_elements(data)

IF savebad THEN BEGIN
   good = where(bad eq 0,count)
   IF count EQ 0 THEN BEGIN
      print, 'All points removed.'
      avg=0.
      avgdev=0.
      stddev=0.
      var=0.
      skew=0.
      kurt=0.
      return
   ENDIF ELSE IF count EQ 1 THEN BEGIN
      avg=data[good[0]]
      avgdev=0.
      stddev=0.
      var=0.
      skew=0.
      kurt=0.
      return
   ENDIF ELSE BEGIN
      Moment4, data[good], avg, avgdev, stddev, var, skew, kurt
      n = count
   ENDELSE
ENDIF ELSE BEGIN
   Moment4, data, avg, avgdev, stddev, var, skew, kurt
   n = N_ELEMENTS( data )
ENDELSE

skewerr = SQRT(6.0/n)
kurterr = SQRT(24.0/n)
if n gt 1 then $
   stdmean = stddev / SQRT(n-1.0) $
else $
   stdmean = stddev

if stddev eq 0 then stddev=max(data)
IF savebad THEN BEGIN
   nogood = where( abs(data[good]-avg)/stddev gt thresh, badcount)
   IF badcount ne 0 THEN bad[good[nogood]]=1
   good = WHERE( bad eq 0, count)
ENDIF ELSE BEGIN
   good = WHERE( ABS( data-avg ) / stddev  LE thresh, count)
ENDELSE

IF count eq 0 THEN BEGIN
   new = data
   nfinal = n_elements(new)
   r_stdmean = stddev
   return
ENDIF

new = data[ good ]
nfinal = n

newlen = count
oldmean = avg + 2.0*eps*stddev ; Sets value high to get things going.
WHILE ( (newlen NE n) AND ( (ABS( oldmean-avg ) / stdmean GT eps) OR $
      ( kurt GT thresh*kurterr ) ) ) DO BEGIN

   oldmean = avg
   n = newlen
   Moment4, new, avg, avgdev, stddev, var, skew, kurt

   if stddev eq 0 then stddev=max(data)

   skewerr = SQRT(6.0/n)
   kurterr = SQRT(24.0/n)
   if n gt 1 then $
      stdmean = stddev / SQRT(n-1.0) $
   else $
      stdmean = stddev

;PRINT, $
;': mean=',avg,' stdmean=',stdmean,' npts=',n,format='(a,g10.4,a,g10.4,a,i9)'

   IF savebad THEN BEGIN
      nogood = where( abs(data[good]-avg)/stddev gt thresh, badcount)
      IF badcount NE 0 THEN bad[good[nogood]]=1
      good = WHERE( bad eq 0, count)
      IF count ne 0 THEN new = data[good]
   ENDIF ELSE BEGIN
      good = WHERE( ABS( new-avg ) / stddev LE thresh, count)
      IF count NE 0 THEN new = new[ good ]
   ENDELSE
   IF count EQ 0 THEN MESSAGE, 'All points removed.'
   newlen = count

ENDWHILE

nfinal = n
r_stdmean = stdmean

END
